from future.standard_library import install_aliases

install_aliases()

import base64
import hashlib
import logging
import re
import socket
import telnetlib
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Union, Any, Optional, Dict, Callable, List, Pattern, Match, Tuple

from .results import BrokenServiceException, OfflineException

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import requests

PORT_MAX = 65535

logging.basicConfig(level=logging.DEBUG)
utilslogger = logging.Logger(__name__)
utilslogger.setLevel(logging.DEBUG)


def assert_in(o1, o2, message=None):
    # type: (Any, Any, Optional[str]) -> None
    """
    Raises an exception if o1 not in o2
    :param o1: the object that should be in o2
    :param o2: the object to look in
    :param message: An optional message that will be part of the error
    """
    if message is None:
        message = "{} is not in {}".format(o1, o2)
    if not o2 or o1 not in o2:
        raise BrokenServiceException(message)


def assert_equals(o1, o2, message=None, autobyteify=False):
    # type: (Any, Any, Optional[str], bool) -> None
    """
    Raises an exception if o1 != o2
    :param o1: the first object
    :param o2: the second object
    :param message: The exception message in case of an error (optional)
    :param autobyteify: will call ensure_bytes on both parameters.
    """
    if message is None:
        message = "{} is not equal to {}".format(o1, o2)
    if autobyteify:
        o1 = ensure_bytes(o1)
        o2 = ensure_bytes(o2)
    if o1 != o2:
        raise BrokenServiceException(message)


def ensure_bytes(obj):
    # type: (Union[bytes, str, Any]) -> bytes
    """
    Converts to bytes
    :param obj: Str or Bytes (or anything else) to convert to bytes representation
    """
    if obj is None:
        raise ValueError("Cannot byteify None")
    if isinstance(obj, bytes):
        return obj
    if isinstance(obj, str):
        return obj.encode("utf-8")
    return str(obj).encode("utf-8")


def ensure_unicode(obj):
    # type: (Union[bytes, str, Any]) -> str
    """Converts to utf-8"""
    if str is None:
        raise ValueError("Cannot stringify None")
    if isinstance(obj, bytes):
        return obj.decode("utf-8")
    if isinstance(obj, str):
        return obj
    return str(obj)


def ensure_valid_filename(s, min_length=3):
    # type: (str, int) -> str
    """
    Gets a valid file name from the input
    :param s: The input string
    :param min_length: if the result is smaller than this, the method will fall back to base64.
    :return: all illegal chars stripped or base64ified if it gets too small
    """
    orig = s
    s = str(s).strip().replace(" ", "_")
    s = re.sub(r"(?u)[^-\w.]", "", s)
    if len(s) < min_length:
        s = base64ify(orig, "+-")
    return s


def snake_caseify(camel):
    # type: (Union[str, bytes]) -> str
    """
    Turn camels into snake (-cases)
    :param camel: camelOrSnakeWhatever
    :return: camel_or_snake_whatever
    """
    camel = ensure_unicode(camel)
    half_snake = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", camel)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", half_snake).lower()


def sha256ify(s):
    # type: (Union[str, bytes]) -> str
    """
    Calculate the sha256 hash
    :param s: the string
    :return: the hash in hex representation
    """
    s = ensure_bytes(s)
    return hashlib.sha256(s).hexdigest()


def base64ify(s, altchars=None):
    # type: (Union[str, bytes], Union[str, bytes, None]) -> str
    """
    Calculate the base64 representation of a value
    :param s: the input string
    :param altchars: base64 encodes using the given altchars (or not, if None)
    :return: base64 representation
    """
    s = ensure_bytes(s)
    if altchars != None:
        altchars = ensure_bytes(altchars)
        return base64.b64encode(s, altchars).decode("utf-8")
    else:
        return base64.b64encode(s).decode("utf-8")


def debase64ify(s, altchars=None):
    # type: (Union[str, bytes], Union[str, bytes, None]) -> str
    """
    Return a string out of a base64
    :param s: the string
    :param altchars: base64 decodes using the given altchars (or not, if None)
    :return: the original value
    """
    s = ensure_bytes(s)
    if altchars != None:
        altchars = ensure_bytes(altchars)
        return base64.b64decode(s, altchars).decode("utf-8")
    else:
        return base64.b64decode(s).decode("utf-8")


def readline_expect(telnet, expected, read_until=b"\n", timeout=30):
    # type: (telnetlib.Telnet, Union[str, bytes], Union[str, bytes], int) -> bytes
    """
    Reads to newline (or read_until string) and assert the presence of a string in the response.
    Will raise an exception if failed.
    :param telnet: Connected telnet instance (the result of self.telnet(..))
    :param expected: the expected String to search for in the response
    :param read_until: Which char to read until.
    :param timeout: a timeout
    :return read: the bytes read
    """
    if hasattr(telnet, "logger"):
        logger = telnet.logger
    else:
        logger = utilslogger

    if isinstance(expected, str):
        expected = expected.encode("utf-8")
    if isinstance(read_until, str):
        read_until = read_until.encode("utf-8")

    read = telnet.read_until(read_until, timeout)
    if read == b"":
        err = "Expected {} but got nothing/timeout!".format(expected)
        logger.error(err, stack_info=True)
        telnet.close()
        raise OfflineException(err)
    if expected not in read:
        err = "Expected {} but got {}".format(expected, read)
        logger.error(err, stack_info=True)
        telnet.close()
        raise BrokenServiceException(err)
    return read


def start_daemon(target):
    # type: (Callable) -> threading.Thread
    """
    starts a thread as daemon
    :param target: the function
    :return: the started thread
    """
    t = threading.Thread(target=target)
    t.daemon = True
    t.start()
    return t


def serve_once(
    html,
    start_port=5000,
    autoincrement_port=True,
    content_type="text/html",
    headers=None,
    logger=None,
):
    # type: (Union[str, bytes, requests.Response], int, bool, str, Optional[Dict[str, str]], Optional[logging.Logger]) -> int
    """
    Render Text in the users browser
    Opens a web server that serves a HTML string once and shuts down after the first request.
    The port will be open when this function returns. (though serving the request may take a few mils)
    :param html: The html code to deliver on the initial request
    :param start_port: The port it should try to listen on first.
    :param autoincrement_port: If the port should be increased if the server cannot listen on the provided start_port
    :param content_type: The content type this server should report (change it if you want json, for example)
    :param headers: Additional headers as {header_key: value} dict.
    :param logger: the optional logger to redirect logs to.
    :return: The port the server started listening on
    """
    # see https://github.com/psf/requests/issues/2925
    import requests

    if logger is None:
        logger = utilslogger
    if headers is None:
        headers = {}
    if isinstance(html, requests.Response):
        html = html.text
    if isinstance(html, str):
        html = html.encode("UTF-8")

    class OutputHandler(BaseHTTPRequestHandler):

        # noinspection PyPep8Naming
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", content_type)
            for key, value in headers.items():
                self.send_header(key, value)
            self.end_headers()
            self.wfile.write(html)
            logger.info("Served HTTP once. Stopping.")
            start_daemon(self.server.shutdown)

    for port in range(start_port, PORT_MAX):
        try:
            server = HTTPServer(("", port), OutputHandler)
            logging.debug("Serving {} bytes on port {}".format(len(html), port))
            start_daemon(server.serve_forever)
            time.sleep(0.1)  # some extra time thrown in for good measure. :)
            return port
        except socket.error as ex:
            if not autoincrement_port:
                logger.info(
                    "Serve once was not set to automatically increment port {} but faced socket exception{}".format(
                        start_port, ex
                    ),
                    exc_info=True,
                    stack_info=True,
                )
                break

    raise socket.error(
        "No unused port found, start_port={}, autoincrement_port={}".format(
            start_port, autoincrement_port
        )
    )


class SimpleSocket(telnetlib.Telnet):
    """
    Telnetlib with some additions.
    Read Telnetlib doku for more.
    """

    # pylint:  disable=protected-access
    def __init__(
        self,
        host=None,
        port=0,
        timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
        logger=None,
        timeout_fun=None,
    ):
        # type: (str, int, int, Optional[logging.Logger], Optional[Callable[[], int]]) -> None
        """
        Initializes a new SimpleSocket Object.
        :param host: the host to connect to
        :param port: the port to connect to
        :param timeout: The timeout passed in here counts for the whole session.
        :param logger: The optional logger to use
        :param timeout_fun: function that will output the current timeout on each call.
        """
        super(SimpleSocket, self).__init__(host, port, timeout)
        self.telnet = super(SimpleSocket, self)  # type: telnetlib.Telnet
        self.socket = self.telnet.get_socket()  # type: socket.socket
        if logger:
            self.logger = logger
        else:
            self.logger = utilslogger
        self.timeout_fun = timeout_fun

    @property
    def current_default_timeout(self):
        # type: () -> int
        """
        Gets the timeout default that should currently be used.
        :return: current timeout default, either from self.timeout_fun or from timeout.
        """
        if self.timeout_fun:
            return self.timeout_fun()
        else:
            return self.timeout

    def readline_expect(self, expected, read_until=b"\n", timeout=None):
        # type: (Union[str, bytes], Union[str, bytes], Optional[int]) -> bytes
        """
        Reads to newline (or read_until string) and assert the presence of a string in the response.
        Will raise an exception if failed.
        :param read_until: Which parameter to read until
        :param expected: the expected String to search for in the response
        :param timeout: The timeout (uses Telnet default if not passed in)
        :return read: the bytes read
        """
        if timeout is None:
            timeout = self.current_default_timeout
        return readline_expect(self, expected, read_until, timeout)

    def expect(self, regexes, timeout=None):
        # type: (List[Union[bytes, Pattern]], Optional[int])->Tuple[int, Match, bytes]
        """
        Read until one from a list of a regular expressions matches.
        Use this to search for anything.

        :param regexes: The first argument is a list of regular expressions, either
        compiled (re.Pattern instances) or uncompiled (strings).
        :param timeout: Timeout in seconds. If none, default will be taken.
        :return: Return a tuple of three items: the index in the list of the
            first regular expression that matches; the re.Match object
            returned; and the text read up till and including the match.
        """
        if timeout is None:
            timeout = self.current_default_timeout

        # Make sure all strings are bytes, ignore compiled Regexes.
        regexes = [ensure_bytes(x) if isinstance(x, str) else x for x in regexes]

        return self.telnet.expect(list=regexes, timeout=timeout)

    def read_until(self, match, timeout=None):
        # type: (Union[bytes, str], Optional[int])->bytes
        """
        Read until the expected string has been seen, or a timeout is hit (default is default socket timeout).
        
        :param match: what to look for.
        :param timeout: default socket timeout override
        :return: Returns everything until the given math. When no match is found, return whatever is available instead,
        possibly the empty string.  Raise EOFError if the connection
        is closed and no cooked data is available.
        """
        if timeout is None:
            timeout = self.current_default_timeout
        return self.telnet.read_until(ensure_bytes(match), timeout)

    def read_n_lines(self, line_count, delimiter=b"\n"):
        # type: (int, Union[str, bytes]) -> List[bytes]
        """
        Read n lines from socket
        :param line_count: the amount of lines to read
        :param delimiter: what delimeter to use for splitting (could also be non-\n)
        :return: a list of lines
        """
        return [self.read_until(ensure_bytes(delimiter)) for _ in range(line_count)]

    def read_all(self):
        # type: () -> bytes
        """
        Read all data until EOF; block until connection closed.
        :return: the complete content until EOF
        """
        return self.telnet.read_all()

    def write(self, buffer):
        # type: (Union[str, bytes]) -> None
        """
        Write a string to the socket.
        Can block if the connection is blocked.
        May raise socket.error if the connection is closed.
        :param buffer: The buffer to write
        """
        self.telnet.write(ensure_bytes(buffer))



import datetime
import json
import logging
import traceback
from logging import LogRecord
from typing import TYPE_CHECKING

LOGGING_PREFIX = "##ENOLOGMESSAGE "

if TYPE_CHECKING:
    from .enochecker import BaseChecker


def exception_to_string(excp):
    stack = traceback.extract_stack()[:-3] + traceback.extract_tb(
        excp.__traceback__
    )  # add limit=??
    pretty = traceback.format_list(stack)
    return "".join(pretty) + "\n  {} {}".format(excp.__class__, excp)


class ELKFormatter(logging.Formatter):
    """
    %(name)s            Name of the logger (logging channel)
    %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
                        WARNING, ERROR, CRITICAL)
    %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
                        "WARNING", "ERROR", "CRITICAL")
    %(pathname)s        Full pathname of the source file where the logging
                        call was issued (if available)
    %(filename)s        Filename portion of pathname
    %(module)s          Module (name portion of filename)
    %(lineno)d          Source line number where the logging call was issued
                        (if available)
    %(funcName)s        Function name
    %(created)f         Time when the LogRecord was created (time.time()
                        return value)
    %(asctime)s         Textual time when the LogRecord was created
    %(msecs)d           Millisecond portion of the creation time
    %(relativeCreated)d Time in milliseconds when the LogRecord was created,
                        relative to the time the logging module was loaded
                        (typically at application startup time)
    %(thread)d          Thread ID (if available)
    %(threadName)s      Thread name (if available)
    %(process)d         Process ID (if available)
    %(message)s         The result of record.getMessage(), computed just as
                        the record is emitted
    """

    def __init__(self, checker, fmt=None, datefmt="%Y-%m-%dT%H:%M:%S%z", style="%"):
        # type: (BaseChecker, str, str, str) -> None
        super().__init__(fmt, datefmt, style)
        self.checker = checker  # type: BaseChecker

    def format(self, record):
        # type: (LogRecord) -> str
        record.stack = self.formatStack(record.stack_info)
        record.asctime = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        # stacktrace = ""
        # if record.exc_info:
        #     print("\n\n\n\n\nlog exc_info:", record.exc_info)
        #     stacktrace = traceback.format_exc(record.exc_info)
        # elif record.stack_info:
        #     stacktrace = record.stack_info
        # if record.exc_info is not None:
        #     print("\n\n\n\n\n\n\n", type(record.exc_info), type(record.stack_info))
        #     print("msg: ", record.exc_info[0], "traceback_info :", record.exc_info[1], "traceback_info :" record.stack_info)

        #     import sys
        #     sys.exit(1)
        if record.exc_info is not None:
            exception_info = {
                "type": record.exc_info[0].__name__,
                "message": str(record.exc_info[1]),
                "traceback": traceback.format_tb(record.exc_info[2], 20),
            }
        else:
            exception_info = None

        log_output = {
            "module": record.module,
            "severity": record.levelname,
            "teamName": self.checker.team,
            "runId": self.checker.run_id,
            "tool": type(self.checker).__name__,
            "flag": self.checker.flag,
            "type": "infrastructure",
            "function": record.funcName,
            "timestamp": record.asctime,
            "round": self.checker.round,
            "relatedRoundId": self.checker.flag_round,
            "flagIndex": self.checker.flag_idx,
            "message": record.getMessage(),
            "exception": exception_info,
            "stacktrace": record.stack_info,
            "serviceName": self.checker.service_name,
        }

        return LOGGING_PREFIX + json.dumps(log_output)


class RestLogHandler(logging.Handler):
    """
    Simple handler class to send Checker logs off to the logging backend Service.
    """

    def __init__(self, checker, level=logging.DEBUG):
        # type: (BaseChecker, int) -> None
        """
        Create a new handler.
        :param checker: The checker to use
        :param level: the Level
        """
        super(RestLogHandler, self).__init__(level)

        # see https://github.com/psf/requests/issues/2925
        import requests

        self.requests = requests

        self.checker = checker  # type: BaseChecker

    def emit(self, record):
        # type: (LogRecord) -> None
        # timestamp = datetime.datetime.fromtimestamp(record.msecs/1000.0).strftime('%Y-%m-%dT%H:%M:%SZ')
        # timestamp = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        # "millis": record.msecs,
        json = {
            "message": record.getMessage(),
            "timestamp": record.asctime,  # Todo: Might not be available everywhere (?)
            "severity": record.levelname,
            "runId": self.checker.run_id,
            "tag": "{}:{}:{}".format(record.name, record.module, record.funcName),
        }
        try:
            r = self.requests.post(self.checker.log_endpoint, json=json)
            if r.status_code != 200:
                print(
                    "Error while logging. Request to {} returned: {}".format(
                        r.status_code, r.text
                    )
                )
        except Exception as ex:
            print("Error while logging: {}".format(ex))



import collections
import json
import logging
import sys
from typing import TYPE_CHECKING, Callable, Type, Any, List, Union, Dict, Tuple

from flask import Flask, Response, request, jsonify

from .enochecker import Result
from .logging import exception_to_string
from .utils import snake_caseify

# from elasticapm.contrib.flask import ElasticAPM

if TYPE_CHECKING:
    from .enochecker import BaseChecker

logging.basicConfig(level=logging.DEBUG)
logger = logging.Logger(__name__)
logger.setLevel(logging.DEBUG)

# ElasticSearch performance monitoring
# apm = ElasticAPM()

Optional = collections.namedtuple("Optional", "key type default")
Required = collections.namedtuple("Required", "key type")

CHECKER_METHODS = [
    "putflag",
    "getflag",
    "putnoise",
    "getnoise",
    "havoc",
    "exploit",
]  # type: List[str]

# The json spec a checker request follows.
spec = [
    Required("method", CHECKER_METHODS),  # method to execute
    Required("address", str),  # address to check
    Optional("runId", int, 0),  # internal ID of this run inside our db
    Optional("team", str, "FakeTeam"),  # the team name
    Optional("teamId", int, 1),  # team ID
    Optional("round", int, 0),  # which tick we are in
    Optional("relatedRoundId", int, 0),  # Flag-Related
    Optional("roundLength", int, 300),  # the default tick time
    Optional("flag", str, "ENOTESTFLAG"),  # the flag or noise to drop or get
    Optional(
        "flagIndex", int, 0
    ),  # the index of this flag in a given round (starts at 0)
    Optional("timeout", int, 30),  # timeout we have for this run
    Optional("logEndpoint", str, None),  # endpoint to send runs to
]  # type: List[Union[Required, Optional]]

UI_TEMPLATE = """

"""

tiny_poster = """
<script>
// To make testing/posting a bit easier, we can do it from the browser here.
var checker_request_count = 0
var checker_pending_requests = 0
var checker_results = []

function post(str) {
    var xhr = new XMLHttpRequest()
    var started = Date.now()
    xhr.open("POST", "/")
    xhr.setRequestHeader("Content-Type", "application/json")
    xhr.onerror = console.error
    xhr.onload = xhr.onerror = function () {
        checker_results = ["Request " + checker_request_count.toString() + " resulted after " + ((Date.now() - started)/1000) + " s in:\\n" + xhr.responseText + "\\n"].concat(checker_results)
        console.log(xhr.responseText)
        document.getElementById("out").innerText = "<plaintext>\\n\\n" + checker_results.join("\\n")
        checker_request_count++
        checker_pending_requests--
        update_pending()
    }
    xhr.send(str)
    checker_pending_requests++
    update_pending()
}

function update_pending(){
    if (checker_pending_requests === 0) {
    	document.getElementById("pending_para").textContent = ""
    } else {
    	document.getElementById("pending_para").textContent = checker_pending_requests.toString() + "Requests pending"
    }	
}

</script>
<div>
<p>Only select one method from the given list.</p>
<p>Values in brackets are optional, so you can delete those lines if you don't want to specify them.</p>
<button onclick=post(document.getElementById("jsonTextbox").value)>Post</button></div>
<p id="pending_para"></p> 
"""


def check_type(name, val, expected_type):
    # type: (str, str, Any) -> None
    """
    returns and converts if necessary
    :param name: the name of the value
    :param val: the value to check
    :param expected_type: the expected type
    """
    if isinstance(expected_type, list):
        if val not in expected_type:
            # The given type is not in the list of allowed members.
            raise ValueError(
                "{} is not a member of expected list {}, got {}".format(
                    name, expected_type, val
                )
            )
    elif not isinstance(val, expected_type):
        raise ValueError(
            "{} should be a '{}' but is of type '{}'.".format(
                name, expected_type.__name__, type(val).__name__
            )
        )


# def generate_form(spec):
#    form = "<form class=\"json-form\">\n"
#    for entry in spec:
#        if isinstance(entry, Required):


def stringify_spec_entry(entry):
    # type: (Union[Optional, Required]) -> str
    """Make a nice string out of it."""
    entrytype = entry.type
    if isinstance(entrytype, type):
        entrytype = entrytype.__name__
    # ugly hack: We don't want a list to be ['like', 'this'] but ["with", "json", "quotes"]...
    entrytype = "{}".format(entrytype).replace("'", '"')
    if isinstance(entry, Required):
        return '"{}": {}'.format(entry.key, entrytype)
    if isinstance(entry, Optional):
        return '"{}": ({} ?? {})'.format(entry.key, entrytype, entry.default)
    raise ValueError(
        "Could not stringify unknown entry type {}: {}".format(type(entry), entry)
    )


def serialize_spec(spec):
    # type: (List[Union[Optional, Required]]) -> str
    """
    Prints a checker json spec in a readable multiline format
    :param spec: a spec
    :return: formatted string
    """
    ret = "{\n"
    for entry in spec:
        if ret != "{\n":
            ret += ",\n"
        ret += "  " + stringify_spec_entry(entry)
    return ret + "\n}"


def assert_types(json, spec):
    # type: (Dict[str, Any], List[Union[Optional, Required]]) -> Dict[str, Any]
    """
    Generates a kwargs dict from a json.
    Will copy all elements from json to the dict, rename all keys to snake_case and Index to idx.
    In case the spec fails, errors out with ValueError.
    :param json:  the json
    :param spec: the spec
    :return: kwargs dict.
    """
    ret = {}

    def key_to_name(key):
        # type: (str)->str
        key = key.replace("Index", "Idx")  # -> flagIndex -> flag_idx
        key = key.replace("relatedRoundId", "flagRound")
        return snake_caseify(key)

    for entry in spec:
        if entry.key not in json:
            if isinstance(entry, Optional):
                ret[key_to_name(entry.key)] = entry.default
            else:
                raise ValueError(
                    "Required parameter {} is missing.".format(
                        stringify_spec_entry(entry)
                    )
                )
        else:

            val = json[entry.key]
            if val is None and isinstance(entry, Optional):
                print("Inserted default")
                val = entry.default

            if entry.key == "method" and val == "havok":
                logger.warning("Ignoring Havok -- calling Havoc instead")
                val = "havoc"
            check_type(entry.key, val, entry.type)
            ret[key_to_name(entry.key)] = val
    return ret


def checker_routes(checker_cls):
    # type: (Type[BaseChecker]) -> Tuple[Callable[[],Response], Callable[[], Response]]
    """
    Creates a flask app for the given checker class.
    :param checker_cls: The checker class to use
    :return: A flask app that can be passed to a uWSGI server or run using .run().
    """

    def index():
        # type: () -> Response
        """
        Some info about this service
        :return: Printable fun..
        """
        logging.info("Request on /")

        return Response(
            "<h1>Welcome to {} :)</h1>"
            '<p>Expecting POST with a JSON:</p><div><textarea id="jsonTextbox" rows={} cols="80">{}</textarea>{}</div>'
            '<a href="https://www.youtube.com/watch?v=SBClImpnfAg"><br>check it out now</a><div id="out">'.format(
                checker_cls.__name__, len(spec) + 3, serialize_spec(spec), tiny_poster
            )
        )

    def serve_checker():
        # type: () -> Response
        """
        Serves a single checker request.
        The spec needs to be formed according to the spec above.
        :param method: the method to run in a checker.
        :return: jsonified result of the checker.
        """
        try:
            logger.info(request.json)
            req_json = request.get_json(force=True)

            kwargs = assert_types(req_json, spec)

            checker = checker_cls(request_dict=kwargs, **kwargs)

            checker.logger.info(request.json)
            res = checker.run().name

            req_json["result"] = res
            req_json = json.dumps(req_json)

            # checker.logger.info("Run resulted in {}: {}".format(res, request.json))
            checker.logger.info("{}".format(req_json))

            return jsonify({"result": res})
        except Exception as ex:
            print(ex)
            logger.error(
                "Returning Internal Error {}.\nTraceback:\n{}".format(
                    ex, exception_to_string(ex)
                ),
                exc_info=ex,
            )
            return jsonify(
                {
                    "result": Result.INTERNAL_ERROR.name,
                    "message": str(ex),
                    "traceback": exception_to_string(ex),
                }
            )

    def service_info():
        # type: () -> Response
        """
        Serves a single checker request.
        The spec needs to be formed according to the spec above.
        :param method: the method to run in a checker.
        :return: jsonified result of the checker.
        """
        try:

            if not hasattr(checker_cls, "service_name"):
                service_name = checker_cls.__name__.split("Checker")[0]
            else:
                service_name = checker_cls.service_name

            info_dict = {
                "serviceName": service_name,
                "flagCount": checker_cls.flag_count,
                "havocCount": checker_cls.havoc_count,
                "noiseCount": checker_cls.noise_count,
            }

            assert isinstance(info_dict["serviceName"], str)
            assert isinstance(info_dict["flagCount"], int)
            assert isinstance(info_dict["havocCount"], int)
            assert isinstance(info_dict["noiseCount"], int)

        except Exception:
            print("SERVICE INFO NOT SPECIFIED!!!11ELF!")
            print(
                "add service_name, flag_count, havoc_count and noise_count as \
static fields to your CHECKER\n"
            )
            print(
                """
Example:
class ExampleChecker(BaseChecker):
    flag_count  = 1
    noise_count = 1
    havoc_count = 1
"""
            )
            raise AttributeError("REQUIRED SERVICE INFO FIELDS NOT SPECIFIED!")

        return info_dict

    def get_service_info():
        return jsonify(service_info())

    return index, serve_checker, service_info, get_service_info


def init_service(checker):
    # type: (Type[BaseChecker]) -> Flask
    """
    Initializes a flask app that can be used for WSGI or listen directly.
    The Engine may Communicate with it over socket.
    :param checker: the checker class to use for check requests.
    :return: a flask app with post and get routes set, ready for checking.
    """
    app = Flask(__name__)
    index, checker_route, service_info, get_service_info = checker_routes(checker)

    app.route("/", methods=["GET"])(index)
    app.route("/", methods=["POST"])(checker_route)
    app.route("/service", methods=["GET"])(get_service_info)

    print(service_info())

    if "run" not in sys.argv:
        # ElasticSearch Performance Monitoring (disabled on commandline)
        # apm.init_app(app, service_name=checker.__name__.split("Checker")[0])  # secret_token=SECRET)
        pass

    return app  # Start service using service.run(host="0.0.0.0")



import datetime
import socket

from flask import Flask
from future.standard_library import install_aliases

install_aliases()

import argparse
import json
import logging
import os
import sys

from abc import ABCMeta, abstractmethod
from typing import Optional, Callable, Any, Dict, List, Union, Type, TYPE_CHECKING
from urllib.parse import urlparse

from future.utils import with_metaclass

from concurrent.futures import TimeoutError

from .utils import snake_caseify, SimpleSocket
from .useragents import random_useragent
from .results import Result, EnoException
from .checkerservice import init_service, CHECKER_METHODS
from .logging import RestLogHandler, ELKFormatter
from .storeddict import StoredDict, DB_DEFAULT_DIR, DB_GLOBAL_CACHE_SETTING
from .nosqldict import NoSqlDict

if TYPE_CHECKING:
    # The import might fail in UWSGI, see the comments below.
    import requests

TIME_BUFFER = 5  # type: int # time in seconds we try to finish earlier

VALID_ARGS = [
    "method",
    "address",
    "team",
    "team_id",
    "round",
    "flag_round",
    "flag",
    "timeout",
    "flag_idx",
    "json_logging",
    "log_endpoint",
    "round_length",
]

# Global cache for all stored dicts.  TODO: Prune this at some point?
global_db_cache = {}  # type: Dict[str, Union[StoredDict, NoSqlDict]]


def parse_args(argv=None):
    # type: (Union[None, List[str], argparse.Namespace]) -> argparse.Namespace
    """
    Returns the parsed argparser args.
    Args look like this:
    [
        "StoreFlag|RetrieveFlag|StoreNoise|RetrieveNoise|Havoc", [Task type]
        "$Address", [Address, either IP or domain]
        "$TeamName",
        "$Round",
        "$Flag|$Noise",
        "$MaxRunningTime",
        "$CallIdx" [index of this task (for each type) in the current round]
    ]
    :param argv: argv. Custom argvs. Will default to sys.argv if not provided.
    :return: args object
    """
    if argv is None:
        return parse_args(sys.argv[1:])
    choices = CHECKER_METHODS + ["listen"]
    parser = argparse.ArgumentParser(description="Your friendly checker script")
    # noinspection SpellCheckingInspection
    subparsers = parser.add_subparsers(
        help="The checker runmode (run/listen)", dest="runmode"
    )
    subparsers.required = True

    listen = subparsers.add_parser("listen", help="Spawn checker service")
    listen.add_argument(
        "listen_port", help="The port the checker service should listen on"
    )

    runparser = subparsers.add_parser("run", help="Run checker on cmdline")
    runparser.add_argument(
        "method",
        choices=choices,
        help='The Method, one of {} or "listen" to start checker service'.format(
            CHECKER_METHODS
        ),
    )
    runparser.add_argument(
        "-a",
        "--address",
        type=str,
        default="localhost",
        help="The ip or address of the remote team to check",
    )
    runparser.add_argument(
        "-t",
        "--team",
        type=str,
        default="team",
        help="The name of the target team to check",
    )
    runparser.add_argument(
        "-T",
        "--team_id",
        type=int,
        default=1,
        help="The Team_id belonging to the specified Team",
    )
    runparser.add_argument(
        "-I",
        "--run_id",
        type=int,
        default=1,
        help="An id for this run. Used to find it in the DB later.",
    )
    runparser.add_argument(
        "-r", "--round", type=int, default=1, help="The round we are in right now"
    )
    runparser.add_argument(
        "-R",
        "--round_length",
        type=int,
        default=300,
        help="The round length in seconds (default 300)",
    )
    runparser.add_argument(
        "-f",
        "--flag",
        type=str,
        default="ENOFLAGENOFLAG=",
        help="The Flag, a Fake flag or a Unique ID, depending on the mode",
    )
    runparser.add_argument(
        "-F",
        "--flag_round",
        type=int,
        default=1,
        help="The Round the Flag belongs to (was placed)",
    )
    runparser.add_argument(
        "-x",
        "--timeout",
        type=int,
        default=30,
        help="The maximum amount of time the script has to execute in seconds",
    )
    runparser.add_argument(
        "-i",
        "--flag_idx",
        type=int,
        default=0,
        help="Unique numerical index per round. Each id only occurs once and is tighly packed, "
        "starting with 0. In a service supporting multiple flags, this would be used to "
        "decide which flag to place.",
    )
    runparser.add_argument(
        "-l",
        "--log_endpoint",
        type=str,
        default="",
        help="URI to an optional RESTlike service accepting log jsons via POST.",
    )
    runparser.add_argument(
        "-j",
        "--json_logging",
        dest="json_logging",
        action="store_true",
        help="If set, logging will be in ELK/Kibana friendly JSON format.",
    )

    return parser.parse_args(args=argv)  # (return is of type argparse.Namespace)


class _CheckerMeta(ABCMeta):
    """
    Some python magic going on right here.
    Each time we subclass BaseChecker, this __init__ is called.
    ABCMeta is used as superclass instead of type, such that BaseChecker is declared abstract -> needs to be overridden.
    """

    def __init__(cls, name, bases, clsdict):
        # type: (Type[_CheckerMeta], str, Dict[Any]) -> None
        """
        Called whenever this class is subclassed.
        :param name: The name of the new class
        :param bases: Bases classes this class inherits from.
        :param clsdict: Contents of this class (.__dict__)
        """
        if len(cls.mro()) > 2:  # 1==BaseChecker
            cls.service = init_service(cls)  # type: Flask
        super(_CheckerMeta, cls).__init__(name, bases, clsdict)


class BaseChecker(with_metaclass(_CheckerMeta, object)):
    """
    All you base are belong to us. Also all your flags. And checker scripts.
    Override the methods given here, then simply init and .run().
    Magic.
    """

    def __init__(
        self,
        request_dict: Dict[str, Any] = None,
        run_id: int = None,
        method: str = None,
        address: str = None,
        team: str = None,
        team_id: int = None,
        round: int = None,
        flag_round: int = None,
        round_length: int = 300,
        flag: str = None,
        flag_idx: int = None,
        timeout: int = None,
        storage_dir: str = DB_DEFAULT_DIR,
        log_endpoint: Optional[str] = None,
        use_db_cache: bool = DB_GLOBAL_CACHE_SETTING,
        json_logging: bool = True,
    ) -> None:
        """
        Inits the Checker, filling the params, according to:
        :param: run_id: Unique ID for this run, assigned by the ctf framework. Used as handle for logging.
        :param: method: The method to run
        """
        # We import requests after startup global imports may deadlock, see
        # https://github.com/psf/requests/issues/2925
        import requests

        self.requests = requests

        self.time_started_at = datetime.datetime.now()  # type: datetime
        self.run_id = run_id  # type: int
        self.log_endpoint = log_endpoint  # type: Optional[str]
        self.json_logging = json_logging  # type: bool

        self.method = method  # type: str
        self.address = address  # type: str
        self.team = team  # type: str
        self.team_id = team_id
        self.round = round  # type: int
        self.current_round = round
        self.flag_round = flag_round  # type: int
        self.round_length = round_length  # type: int
        self.flag = flag  # type: str
        self.timeout = timeout  # type: int

        self.flag_idx = flag_idx  # type: int
        self.storage_dir = storage_dir

        self._setup_logger()
        if use_db_cache:
            self._active_dbs = global_db_cache  # type: Dict[str, StoredDict]
        else:
            self._active_dbs = {}  # type: Dict[str, Union[NoSqlDict, StoredDict]]
        self.http_session = self.requests.session()  # type: requests.Session
        self.http_useragent = random_useragent()

        if not hasattr(self, "service_name"):
            self.service_name = type(self).__name__.split("Checker")[0]
            self.debug(
                "Assuming checker Name {}. If that's not the case, please override.".format(
                    self.service_name
                )
            )

        if not hasattr(self, "port"):
            self.warning("No default port defined.")
            self.port = -1

        self.request_dict = request_dict  # kinda duplicate
        self.config = {x: getattr(self, x) for x in VALID_ARGS}

        self.debug(
            "Initialized checker for flag {} with in {} seconds".format(
                json.dumps(self.config), datetime.datetime.now() - self.time_started_at
            )
        )

        if self.method == "havok":
            self.method = "havoc"
            self.warning("Ignoring method 'havok', calling 'havoc' instead")

    def _setup_logger(self):
        # type: () -> None
        """
        Sets up a logger usable from inside a checker using
        self.debug, self.info, self.warning, self.error or self.logger
        A logger can have additional args as well as exc_info=ex to log an exception, stack_info=True to log trace.
        """
        self.logger = logging.Logger(type(self).__name__)  # type: logging.Logger
        self.logger.setLevel(logging.DEBUG)

        # default handler
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        if self.json_logging:
            formatter = ELKFormatter(self)
        else:
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        if self.log_endpoint and self.log_endpoint.startswith("http"):
            self.logger.addHandler(RestLogHandler(self))

        self.debug = self.logger.debug  # type: Callable[[str, ...], None]
        self.info = self.logger.info  # type: Callable[[str, ...], None]
        self.warning = self.logger.warning  # type: Callable[[str, ...], None]
        self.error = self.logger.error  # type: Callable[[str, ...], None]

    @property
    def noise(self):
        # type: () -> str
        """
        Pretty similar to a flag, just in a different mode (storeNoise vs storeFlag)
        :return: The noise
        """
        return self.flag

    @property
    def time_running(self):
        # type: () -> float
        """
        How long this checker has been running
        :return: time this checker has been running for
        """
        return (datetime.datetime.now() - self.time_started_at).total_seconds()

    @property
    def time_remaining(self):
        # type: () -> int
        """
        Returns a remaining time that is save to be used as timeout (includes a buffer of TIME_BUFFER seconds)
        :return: A save number of seconds that may still be used
        """
        return max(int(self.timeout - self.time_running - TIME_BUFFER), 1)

    # def __format_internal_db_entry(name):
    #     return f"__Checker-Internals:{name}__"

    # ---- Basic checker functionality ---- #

    def run(self, method=None):
        # type: (Optional[str, Callable]) -> Result
        """
        Executes the checker and catches errors along the way.
        :param method: When calling run, you may call a different method than the one passed on Checker creation
                        using this optional param.
        :return: the Result code as int, as per the Result enum.
        """

        try:
            if callable(method):
                ret = method()
            else:
                if method is None:
                    method = self.method
                if method not in CHECKER_METHODS:
                    raise ValueError(
                        "Method {} not supported! Supported: {}".format(
                            method, CHECKER_METHODS
                        )
                    )

                ignore_run = False
                if method == "getflag":
                    try:
                        ignore_run = not (
                            "OK"
                            == self.team_db[
                                f"__Checker-internals-RESULT:putflag,{self.flag_round},{self.flag_idx}__"
                            ]
                        )

                    except KeyError as ex:
                        self.info(
                            f"original putflag did not return successfully -- ignoring getflag for flag_round:{self.flag_round}, index: {self.flag_idx}"
                        )
                        ignore_run = True

                if method == "getnoise":
                    try:

                        ignore_run = not (
                            "OK"
                            == self.team_db[
                                f"__Checker-internals-RESULT:putnoise,{self.flag_round},{self.flag_idx}__"
                            ]
                        )

                    except KeyError as ex:
                        self.info(
                            f"original putnoise did not return successfully -- ignoring getnoise for flag_round:{self.flag_round}, index: {self.flag_idx}"
                        )
                        ignore_run = True

                if ignore_run:
                    self.debug("run ignored -- preemptively returned OK")
                    return Result.OK

                ret = getattr(self, snake_caseify(method))()
            if Result.is_valid(ret):
                ret = Result(
                    ret
                )  # Better wrap this, in case somebody returns raw ints (?)
                self.info("Checker [{}] resulted in {}".format(self.method, ret.name))
                self.team_db[
                    f"__Checker-internals-RESULT:{str(method)},{self.flag_round},{self.flag_idx}__"
                ] = ret.name
                return ret
            if ret is not None:
                self.error("Illegal return value from {}: {}".format(self.method, ret))
                return (
                    Result.INTERNAL_ERROR
                )  # , "Illegal return value from {}: {}".format(self.method, ret)

            # Returned Normally
            self.info("Checker [{}] executed successfully!".format(self.method))
            self.team_db[
                f"__Checker-internals-RESULT:{str(method)},{self.flag_round},{self.flag_idx}__"
            ] = "OK"
            return Result.OK

        except EnoException as eno:
            self.info(
                "Checker[{}] result: {}({})".format(self.method, eno.result.name, eno),
                exc_info=eno,
            )
            return Result(eno.result)  # , eno.message
        except self.requests.HTTPError as ex:
            self.info("Service returned HTTP Errorcode [{}].".format(ex), exc_info=ex)
            return Result.MUMBLE  # , "HTTP Error" #For now
        except (
            self.requests.ConnectionError,  # requests
            self.requests.ConnectTimeout,  # requests
            TimeoutError,
            socket.timeout,
            ConnectionError,
            socket.error,
            # ConnectionAbortedError,  # not in py2, already handled by ConnectionError.
            # ConnectionRefusedError
        ) as ex:
            self.info(
                "Error in connection to service occurred: {}\n".format(ex), exc_info=ex
            )
            return Result.OFFLINE  # , ex.message
        except Exception as ex:
            self.error("Unhandled checker error occurred: {}\n".format(ex), exc_info=ex)
            return Result.INTERNAL_ERROR  # , ex.message
        finally:
            for db in self._active_dbs.values():
                # A bit of cleanup :)
                db.persist()

    @abstractmethod
    def putflag(self):
        # type: () -> Optional[Result]
        """
        This method stores a flag in the service.
        In case multiple flags are provided, self.flag_idx gives the appropriate index.
        The flag itself can be retrieved from self.flag.
        On error, raise an Eno Exception.
        :raises EnoException on error
        :return this function can return a result if it wants
                if nothing is returned, the service status is considered okay.
                the preferred way to report errors in the service is by raising an appropriate enoexception
        """
        pass

    @abstractmethod
    def getflag(self):
        # type: () -> Optional[Result]
        """
        This method retrieves a flag from the service.
        Use self.flag to get the flag that needs to be recovered and self.roudn to get the round the flag was placed in.
        On error, raise an EnoException.
        :raises EnoException on error
        :return this function can return a result if it wants
                if nothing is returned, the service status is considered okay.
                the preferred way to report errors in the service is by raising an appropriate enoexception
        """
        pass

    @abstractmethod
    def putnoise(self):
        # type: () -> Optional[Result]
        """
        This method stores noise in the service. The noise should later be recoverable.
        The difference between noise and flag is that noise does not have to remain secret for other teams.
        This method can be called many times per round. Check how often using self.flag_idx.
        On error, raise an EnoException.
        :raises EnoException on error
        :return this function can return a result if it wants
                if nothing is returned, the service status is considered okay.
                the preferred way to report errors in the service is by raising an appropriate enoexception
        """
        pass

    @abstractmethod
    def getnoise(self):
        # type: () -> Optional[Result]
        """
        This method retrieves noise in the service.
        The noise to be retrieved is inside self.flag
        The difference between noise and flag is, tht noise does not have to remain secret for other teams.
        This method can be called many times per round. Check how often using flag_idx.
        On error, raise an EnoException.
        :raises EnoException on error
        :return this function can return a result if it wants
                if nothing is returned, the service status is considered okay.
                the preferred way to report errors in the service is by raising an appropriate enoexception
        """
        pass

    @abstractmethod
    def havoc(self):
        # type: () -> Optional[Result]
        """
        This method unleashes havoc on the app -> Do whatever you must to prove the service still works. Or not.
        On error, raise an EnoException.
        :raises EnoException on Error
        :return This function can return a result if it wants
                If nothing is returned, the service status is considered okay.
                The preferred way to report Errors in the service is by raising an appropriate EnoException
        """
        pass

    @abstractmethod
    def exploit(self):
        # type: () -> Optional[Result]
        """
        This method is strictly for testing purposes and will hopefully not be called during the actual CTF.
        :raises EnoException on Error
        :return This function can return a result if it wants
                If nothing is returned, the service status is considered okay.
                The preferred way to report Errors in the service is by raising an appropriate EnoException
        """
        pass

    # ---- DB specific methods ---- #
    def db(self, name, ignore_locks=False):
        # type: (str, bool) -> Union[NoSqlDict, StoredDict]
        """
        Get a (global) db by name
        Subsequent calls will return the same db.
        Names can be anything, for example the team name, round numbers etc.
        :param name: The name of the DB
        :param ignore_locks: Should only be set if you're sure-ish keys are never shared between instances.
                Manual locking ist still possible.
        :return: A dict that will be self storing. Alternatively,
        """
        try:
            db = self._active_dbs[name]
            db.logger = self.logger
            # TODO: Settng a new Logger backend may throw logs in the wrong direction in a multithreaded environment!
            return db
        except KeyError:
            checker_name = type(self).__name__
            self.debug("Remote DB {} was not cached.".format(name))
            if os.getenv("MONGO_ENABLED"):
                host = os.getenv("MONGO_HOST")
                port = int(os.getenv("MONGO_PORT"))
                username = os.getenv("MONGO_USER")
                password = os.getenv("MONGO_PASSWORD")
                self.debug(
                    "Using NoSqlDict mongo://{}:{}@{}:{}".format(
                        username, "".join(["X" for c in password]), host, port
                    )
                )

                ret = NoSqlDict(
                    name=name,
                    checker_name=checker_name,
                    host=host,
                    port=port,
                    username=username,
                    password=password,
                )
            else:
                self.debug(
                    "MONGO_ENABLED not set. Using stored dict at {}".format(
                        self.storage_dir
                    )
                )
                ret = StoredDict(
                    name=name,
                    base_path=self.storage_dir,
                    ignore_locks=ignore_locks,
                    logger=self.logger,
                )
            self._active_dbs[name] = ret
            return ret

    @property
    def global_db(self):
        # type: () -> StoredDict
        """
        A global storage shared between all teams and rounds.
        Subsequent calls will return the same db.
        Prefer db_team_local or db_round_local
        :return: The global db
        """
        return self.db("global")

    @property
    def team_db(self):
        # type: () -> StoredDict
        """
        The database for the current team
        :return: The team local db
        """
        return self.get_team_db()

    def get_team_db(self, team=None):
        # type: (Optional[str]) -> StoredDict
        """
        Returns the database for a specific team.
        Subsequent calls will return the same db.
        :param team: Return a db for an other team. If none, the db for the local team will be returned.
        :return: The team local db
        """
        team = team if team is not None else self.team
        return self.db("team_{}".format(team), ignore_locks=True)

    # ---- Networking specific methods ---- #
    def _sanitize_url(self, route, port=None, scheme=None):
        # type: (str, Optional[int], Optional[str]) -> str
        if port is None:
            port = self.port
        if port is None:
            raise ValueError("Port for service not set. Cannot Request.")

        if ":" in self.address:
            netloc = "[{}]:{}".format(self.address, port)
        else:
            netloc = "{}:{}".format(self.address, port)
        if scheme is None:
            url = urlparse(route)
        else:
            url = urlparse(route, scheme=scheme)
        # noinspection PyProtectedMember
        return url._replace(netloc=netloc).geturl()

    def connect(self, host=None, port=None, timeout=None):
        # type: (Optional[str], Optional[int], Optional[int]) -> SimpleSocket
        """
        Opens a socket/telnet connection to the remote host.
        Use connect(..).get_socket() for the raw socket.
        :param host: the host to connect to (defaults to self.address)
        :param port: the port to connect to (defaults to self.port)
        :param timeout: timeout on connection (defaults to self.timeout)
        :return: A connected Telnet instance
        """
        timeout_fun = lambda: timeout
        if timeout is None:
            timeout = self.time_remaining / 2
            timeout_fun = lambda: self.time_remaining / 2
        if port is None:
            port = self.port
        if host is None:
            host = self.address
        self.debug(
            "Opening socket to {}:{} (timeout {} secs).".format(host, port, timeout)
        )
        return SimpleSocket(
            host, port, timeout=timeout, logger=self.logger, timeout_fun=timeout_fun
        )

    @property
    def http_useragent(self):
        # type: () -> str
        """
        The useragent for http(s) requests
        :return: the current useragent
        """
        return self.http_session.headers["User-Agent"]

    @http_useragent.setter
    def http_useragent(self, useragent):
        # type: (str) -> None
        """
        Sets the useragent for http requests.
        Randomize using http_useragent_randomize()
        :param useragent: the useragent
        """
        self.http_session.headers["User-Agent"] = useragent

    def http_useragent_randomize(self):
        """
        Choses a new random http useragent.
        Note that http requests will be initialized with a random user agent already.
        To retrieve a random useragent without setting it, use random instead.
        :return: the new useragent
        """
        new_agent = random_useragent()
        self.http_useragent = new_agent
        return new_agent

    def http_post(
        self,
        route="/",
        params=None,
        port=None,
        scheme="http",
        raise_http_errors=False,
        timeout=None,
        **kwargs,
    ):
        # type: (str, Any, Optional[int], str, bool, Optional[int], ...) -> "requests.Response"
        """
        Performs a (http) requests.post to the current host.
        Caches cookies in self.http_session
        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        return self.http(
            "post", route, params, port, scheme, raise_http_errors, timeout, **kwargs
        )

    def http_get(
        self,
        route="/",
        params=None,
        port=None,
        scheme="http",
        raise_http_errors=False,
        timeout=None,
        **kwargs,
    ):
        # type: (str, Any, Optional[int], str, bool, Optional[int], ...) -> "requests.Response"
        """
        Performs a (http) requests.get to the current host.
        Caches cookies in self.http_session
        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        return self.http(
            "get", route, params, port, scheme, raise_http_errors, timeout, **kwargs
        )

    def http(
        self,
        method,
        route="/",
        params=None,
        port=None,
        scheme="http",
        raise_http_errors=False,
        timeout=None,
        **kwargs,
    ):
        # type: (str, str, Any, Optional[int], str, bool, Optional[int], ...) -> "requests.Response"
        """
        Performs an http request (requests lib) to the current host.
        Caches cookies in self.http_session
        :param method: The request method
        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect (default: self.timeout)
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        url = self._sanitize_url(route, port, scheme)
        if timeout is None:
            timeout = self.time_remaining / 2
        self.debug(
            "Request: {} {} with params: {} and {} secs timeout.".format(
                method, url, params, timeout
            )
        )
        resp = self.http_session.request(
            method, url, params=params, timeout=timeout, **kwargs
        )
        if raise_http_errors:
            resp.raise_for_status()
        return resp


def run(checker_cls, args=None):
    # type: (Type[BaseChecker], Union[None, List[str], argparse.Namespace]) -> None
    """
    # Runs a checker, either from cmdline or as uwsgi script.
    :param checker: The checker (subclass of basechecker) to run
    :param force_service: if True (non-default), the server will skip arg parsing and immediately spawn the web service.
    :param args: optional parameter, providing parameters
    :return:  Never returns.
    """
    parsed = parse_args(args)
    if parsed.runmode == "listen":
        checker_cls.service.run(host="::", debug=True, port=parsed.listen_port)
    else:
        checker_args = vars(parsed)
        del checker_args["runmode"]  # will always be 'run' at this point
        result = checker_cls(**vars(parsed)).run()
        print("Checker run resulted in Result.{}".format(result.name))
        return result



from collections.abc import MutableMapping
import logging
import os
from functools import wraps
from threading import RLock, current_thread
from typing import Iterable, Any, Optional, TYPE_CHECKING, Union

from enochecker import utils
from .utils import base64ify

if TYPE_CHECKING:
    # We do these things as late as possible to avoid strange deadlocks when forking in UWSGI
    from pymongo import MongoClient

# LOGGING SETUP
logging.basicConfig(level=logging.DEBUG)
dictlogger = logging.Logger(__name__)
dictlogger.setLevel(logging.DEBUG)

RETRY_COUNT = 4

# DB DEFAULT PARAMS
DB_DEFAULT_USER = None
DB_DEFAULT_PASS = None
DB_DEFAULT_HOST = "localhost"
DB_DEFAULT_PORT = 27017


def to_keyfmt(key):
    return str(key)  # + type(key).__name__


def _try_n_times(func):
    @wraps(func)
    def try_n_times(*args, **kwargs):
        from pymongo.errors import PyMongoError

        for i in range(RETRY_COUNT):
            try:
                return func(*args, **kwargs)
            except PyMongoError as ex:
                dictlogger.error("noSQLdict_Error, Try {}".format(str(i)), exc_info=ex)
                if i == RETRY_COUNT:
                    raise

    return try_n_times


class NoSqlDict(MutableMapping):
    """
    A dictionary that is MongoDb backed.
    """

    dblock = RLock()

    @classmethod
    def get_client(
        cls, host: str, port: int, username: Optional[str], password: Optional[str]
    ) -> "MongoClient":
        """
        Lazily tries to get the mongo db connection or creates a new one.
        :param host: mongo host
        :param port: mongo port
        :param username: the username to connect to
        :param password: the password to use
        :return:
        """
        mongo_name = utils.ensure_valid_filename(
            "mongo_{}_{}_{}_{}".format(host, port, username, password)
        )
        if hasattr(cls, mongo_name):
            return getattr(cls, mongo_name)

        from pymongo import MongoClient

        with NoSqlDict.dblock:
            if hasattr(cls, mongo_name):
                # we found a mongo in the meantime
                return getattr(cls, mongo_name)
            mongo = MongoClient(
                host=host, port=port, username=username, password=password
            )
            setattr(cls, mongo_name, mongo)
        dictlogger.debug(
            "MONGO CLIENT INITIALIZED for thread {}: {}".format(current_thread(), mongo)
        )
        return mongo

    def __init__(
        self,
        name: str = "default",
        checker_name: str = "BaseChecker",
        host: str = DB_DEFAULT_HOST,
        port: int = DB_DEFAULT_PORT,
        username: str = DB_DEFAULT_USER,
        password: str = DB_DEFAULT_PASS,
        *args,
        **kwargs
    ):
        self.dict_name = base64ify(name, altchars=b"-_")
        self.checker_name = checker_name
        self.cache = {}
        self.db = self.get_client(host, port, username, password)[checker_name][
            self.dict_name
        ]
        try:
            self.db.index_information()["checker_key"]
        except KeyError:
            self.db.create_index(
                [("key", 1)], name="checker_key", unique=True, background=True
            )

    @_try_n_times
    def __setitem__(self, key, value):
        self.cache[key] = value

        query_dict = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }

        to_insert = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
            "value": value,
        }

        self.db.replace_one(query_dict, to_insert, upsert=True)

    @_try_n_times
    def __getitem__(self, key, print_result=False):
        if key in self.cache.items():
            return self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }

        result = self.db.find_one(to_extract)

        if print_result:
            dictlogger.debug(result)

        if result:
            self.cache[key] = result["value"]
            return result["value"]
        raise KeyError("Could not find {} in {}".format(key, self))

    @_try_n_times
    def __delitem__(self, key):
        if key in self.cache:
            del self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }
        self.db.delete_one(to_extract)

    @_try_n_times
    def __len__(self) -> int:
        return self.db.count_documents(
            {"checker": self.checker_name, "name": self.dict_name}
        )

    @_try_n_times
    def __iter__(self) -> Iterable[Any]:
        iterdict = {"checker": self.checker_name, "name": self.dict_name}
        results = self.db.find(iterdict)
        for key in map(lambda res: res["key"], results):
            yield key

    def persist(self) -> None:
        # TODO: could wait until here before hitting the mongodb...
        pass



from future.standard_library import install_aliases

install_aliases()

from .utils import (
    assert_equals,
    assert_in,
    base64ify,
    debase64ify,
    sha256ify,
    ensure_bytes,
    ensure_unicode,
    ensure_valid_filename,
    snake_caseify,
    readline_expect,
    serve_once,
    start_daemon,
    SimpleSocket,
)  # the util stuff
from .results import (
    BrokenServiceException,
    EnoException,
    OfflineException,
    Result,
)  # Possible results
from .enochecker import run, parse_args, BaseChecker  # the BaseChecker
from .checkerservice import CHECKER_METHODS

name = "enochecker"



from random import sample
from typing import List

# Useragent list from ructf github :)
useragents = [
    r"Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; WOW64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Safari/602.1.50",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
    r"Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/602.2.14 (KHTML, like Gecko) Version/10.0.1 Safari/602.2.14",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10.11; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Safari/602.1.50",
    r"Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.79 Safari/537.36 Edge/14.14393",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.3; WOW64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 6.1; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64; rv:45.0) Gecko/20100101 Firefox/45.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/53.0.2785.143 Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/601.7.7 (KHTML, like Gecko) Version/9.1.2 Safari/601.7.7",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/601.7.8 (KHTML, like Gecko) Version/9.1.3 Safari/601.7.8",
    r"Mozilla/5.0 (iPad; CPU OS 10_0_2 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A456 Safari/602.1",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Safari/602.1.50",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0;  Trident/5.0)",
    r"Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0;  Trident/5.0)",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/602.2.14 (KHTML, like Gecko) Version/10.0.1 Safari/602.2.14",
    r"Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64; rv:45.0) Gecko/20100101 Firefox/45.0",
    r"Mozilla/5.0 (Windows NT 5.1; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0 Safari/537.36 Edge/13.10586",
    r"Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64; rv:47.0) Gecko/20100101 Firefox/47.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko",
    r"Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64; rv:48.0) Gecko/20100101 Firefox/48.0",
    r"Mozilla/5.0 (iPhone; CPU iPhone OS 10_0_2 like Mac OS X) AppleWebKit/601.1 (KHTML, like Gecko) CriOS/53.0.2785.109 Mobile/14A456 Safari/601.1.46",
    r"Mozilla/5.0 (iPhone; CPU iPhone OS 10_0_2 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14A456 Safari/602.1",
    r"Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:47.0) Gecko/20100101 Firefox/47.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/601.6.17 (KHTML, like Gecko) Version/9.1.1 Safari/601.6.17",
    r"Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:48.0) Gecko/20100101 Firefox/48.0",
    r"Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36",
    r"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/52.0.2743.116 Chrome/52.0.2743.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.143 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:49.0) Gecko/20100101 Firefox/49.0",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4) AppleWebKit/601.5.17 (KHTML, like Gecko) Version/9.1 Safari/601.5.17",
    r"Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/53.0.2785.116 Safari/537.36",
    r"Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.59 Safari/537.36",
]  # type: List[str]


def random_useragent():
    # type: ()-> str
    """
    Returns a random useragent
    :return: A seemingly valid useragent.
    """
    return sample(useragents, 1)[0]
